from doodl.doodl import \
    areachart, barchart, bollinger, boxplot, bubblechart, chord, \
    contour, dendrogram, disjoint, dotplot, force, gantt, heatmap, \
    linechart, piechart, scatterplot, skey, stacked_areachart, \
    stacked_barchart, tree, treemap, vennchart, voronoi

from doodl.data import interpret_data
